@@ -7,24 +7,51 @@ import warnings
 import traceback
 import os.path
 import multiprocessing
+import numpy as np
 import pysam
 
 import HTSeq
+from HTSeq.scripts.utils import (
+    UnknownChrom,
+    my_showwarning,
+    invert_strand,
+    _write_output,
+)
 
 
-class UnknownChrom(Exception):
-    pass
+def correct_barcodes(counts, hamming=1):
+    '''Correct barcodes, usually UMIs.
 
+    Notice: This function does not use recursive correction. Recursion sounds
+    great in theory, but due to experimental corner cases it can lead to
+    overcorrection and loss of signal.
+    '''
+    if hamming == 0:
+        return
 
-def invert_strand(iv):
-    iv2 = iv.copy()
-    if iv2.strand == "+":
-        iv2.strand = "-"
-    elif iv2.strand == "-":
-        iv2.strand = "+"
-    else:
-        raise ValueError("Illegal strand")
-    return iv2
+    # Count reads from all feature per barcode, and prepare to oder
+    n_reads = Counter({key: sum(val.values()) for key, val in counts.items()})
+
+    # Order by counts, from most to least
+    order = [umi for (umi, _) in n_reads.most_common()]
+
+    # Get close Hamming distances, and aggregate them into higher count ones
+    umi_vectors = np.array([list(x) for x in order])
+    idx_left = list(range(len(order)))
+    while idx_left:
+        i = idx_left.pop(0)
+        vi = ''.join(umi_vectors[i])
+        # Distance from all remaining barcodes
+        dis = (umi_vectors[i] != umi_vectors[idx_left]).sum(axis=1)
+        # Get indices of barcodes within reach
+        idx = (dis <= hamming).nonzero()[0]
+        js = [idx_left[idxi] for idxi in idx]
+        for j in js:
+            # Merge barcode counts into higher count one
+            vj = ''.join(umi_vectors[j])
+            counts[vi].update(counts.pop(vj))
+        # Shorten list of remaining UMIs
+        idx_left = [j for j in idx_left if j not in js]
 
 
 def count_reads_with_barcodes(
@@ -47,6 +74,7 @@ def count_reads_with_barcodes(
         samout_filename,
         cb_tag,
         ub_tag,
+        correct_ub_distance,
         ):
 
     def write_to_samout(r, assignment, samoutfile, template=None):
@@ -57,10 +85,14 @@ def count_reads_with_barcodes(
         for read in r:
             if read is not None:
                 read.optional_fields.append(('XF', assignment))
-                if samout_format in ('SAM', 'sam'):
+                if template is not None:
+                    samoutfile.write(read.to_pysam_AlignedSegment(template))
+                elif samout_format in ('SAM', 'sam'):
                     samoutfile.write(read.get_sam_line() + "\n")
                 else:
-                    samoutfile.write(read.to_pysam_AlignedSegment(template))
+                    raise ValueError(
+                        'BAM/SAM output: no template and not a test SAM file',
+                    )
 
     def identify_barcodes(r):
         '''Identify barcode from the read or pair (both must have the same)'''
@@ -100,6 +132,13 @@ def count_reads_with_barcodes(
                     samout_filename, 'wb',
                     template=template,
                     )
+        elif (samout_format in ('sam', 'SAM')) and \
+                hasattr(read_seq_file, 'get_template'):
+            template = read_seq_file.get_template()
+            samoutfile = pysam.AlignmentFile(
+                    samout_filename, 'w',
+                    template=template,
+                    )
         else:
             template = None
             samoutfile = open(samout_filename, 'w')
@@ -326,11 +365,17 @@ def count_reads_with_barcodes(
     if samoutfile is not None:
         samoutfile.close()
 
-    # Get rid of UMI by majority rule
+    # A UMI could be mapped to more than one feature. We count the feature
+    # with the highest number of reads. In case of ties, we discard the whole
+    # UMI to be on the safe side (it should not happen anyway).
     cbs = sorted(counts.keys())
     counts_noumi = {}
     for cb in cbs:
         counts_cell = Counter()
+
+        # Correct barcodes within a certain Hamming distance
+        correct_barcodes(counts[cb], hamming=correct_ub_distance)
+
         for ub, udic in counts.pop(cb).items():
             # In case of a tie, do not increment either feature
             top = udic.most_common(2)
@@ -358,14 +403,17 @@ def count_reads_in_features(
         feature_type,
         id_attribute,
         additional_attributes,
+        add_chromosome_info,
         quiet,
         minaqual,
         samout,
         samout_format,
         output_delimiter,
         output_filename,
+        counts_output_sparse,
         cb_tag,
         ub_tag,
+        correct_ub_distance,
         ):
     '''Count reads in features, parallelizing by file'''
 
@@ -395,6 +443,7 @@ def count_reads_in_features(
         additional_attributes=additional_attributes,
         stranded=stranded != 'no',
         verbose=not quiet,
+        add_chromosome_info=add_chromosome_info,
         )
     features = feature_scan['features']
     attributes = feature_scan['attributes']
@@ -425,56 +474,21 @@ def count_reads_in_features(
         samout,
         cb_tag,
         ub_tag,
+        correct_ub_distance,
         )
-    # Cell barcodes
-    cbs = results['cell_barcodes']
-    counts = results['counts']
 
     # Write output
-    other_features = [
-        '__no_feature',
-        '__ambiguous',
-        '__too_low_aQual',
-        '__not_aligned',
-        '__alignment_not_unique',
-        ]
-    pad = ['' for attr in additional_attributes]
-    # Header
-    fields = [''] + pad + cbs
-    line = output_delimiter.join(fields)
-    if output_filename == '':
-        print(line)
-    else:
-        with open(output_filename, 'w') as f:
-            f.write(line)
-            f.write('\n')
-
-    # Features
-    for ifn, fn in enumerate(feature_attr):
-        fields = [fn] + attributes[fn] + [str(counts[cb][fn]) for cb in cbs]
-        line = output_delimiter.join(fields)
-        if output_filename == '':
-            print(line)
-        else:
-            with open(output_filename, 'a') as f:
-                f.write(line)
-                f.write('\n')
-
-    # Other features
-    for fn in other_features:
-        fields = [fn] + pad + [str(counts[cb][fn]) for cb in cbs]
-        line = output_delimiter.join(fields)
-        if output_filename == '':
-            print(line)
-        else:
-            with open(output_filename, 'a') as f:
-                f.write(line)
-                f.write('\n')
-
-
-def my_showwarning(message, category, filename, lineno=None, file=None,
-                   line=None):
-    sys.stderr.write("Warning: %s\n" % message)
+    _write_output(
+        results,
+        results['cell_barcodes'],
+        attributes,
+        additional_attributes,
+        output_filename,
+        output_delimiter,
+        False,
+        sparse=counts_output_sparse,
+        dtype=np.float32,
+    )
 
 
 def main():
@@ -491,6 +505,15 @@ def main():
         "Released under the terms of the GNU General Public License v3. " +
         "Part of the 'HTSeq' framework, version %s." % HTSeq.__version__)
 
+    pa.add_argument(
+            "--version", action="store_true",
+            help='Show software version and exit')
+    args, argv = pa.parse_known_args()
+    # Version is the only case where the BAM and GTF files are optional
+    if args.version:
+        print(HTSeq.__version__)
+        sys.exit()
+
     pa.add_argument(
             "samfilename", type=str,
             help="Path to the SAM/BAM file containing the barcoded, mapped " +
@@ -556,6 +579,12 @@ def main():
             "suitable for Ensembl GTF files: gene_name). Use multiple times " +
             "for each different attribute")
 
+    pa.add_argument(
+            "--add-chromosome-info", action='store_true',
+            help="Store information about the chromosome of each feature as " +
+            "an additional attribute (e.g. colunm in the TSV output file).",
+            )
+
     pa.add_argument(
             "-m", "--mode", dest="mode",
             choices=("union", "intersection-strict", "intersection-nonempty"),
@@ -604,6 +633,11 @@ def main():
             help="TSV/CSV filename to output the counts to instead of stdout."
             )
 
+    pa.add_argument(
+            "--counts_output_sparse", action='store_true',
+            help="Store the counts as a sparse matrix (mtx, h5ad, loom)."
+            )
+
     pa.add_argument(
             '--cell-barcode', type=str, dest='cb_tag',
             default='CB',
@@ -615,24 +649,28 @@ def main():
             '--UMI', type=str, dest='ub_tag',
             default='UB',
             help='BAM tag used for the unique molecular identifier, also ' +
-            ' known as molecular barcode (default compatible ' +
+            'known as molecular barcode (default compatible ' +
             'with 10X Genomics Chromium is UB).',
             )
 
+    pa.add_argument(
+            '--correct-UMI-distance',
+            type=int,
+            choices=[0, 1, 2],
+            dest='correct_ub_distance',
+            default=0,
+            help='Correct for sequencing errors in the UMI tag, based on ' +
+            'Hamming distance. For each UMI, if another UMI with more reads ' +
+            'within 1 or 2 mutations is found, merge this UMI\'s reads into ' +
+            'the more popular one. The default is to not correct UMIs.',
+    )
+
     pa.add_argument(
             "-q", "--quiet", action="store_true", dest="quiet",
             help="Suppress progress report")  # and warnings" )
 
-    pa.add_argument(
-            "--version", action="store_true",
-            help='Show software version and exit')
-
     args = pa.parse_args()
 
-    if args.version:
-        print(HTSeq.__version__)
-        sys.exit()
-
     warnings.showwarning = my_showwarning
     try:
         count_reads_in_features(
@@ -648,14 +686,17 @@ def main():
                 args.featuretype,
                 args.idattr,
                 args.additional_attr,
+                args.add_chromosome_info,
                 args.quiet,
                 args.minaqual,
                 args.samout,
                 args.samout_format,
                 args.output_delimiter,
                 args.output_filename,
+                args.counts_output_sparse,
                 args.cb_tag,
                 args.ub_tag,
+                args.correct_ub_distance,
                 )
     except:
         sys.stderr.write("  %s\n" % str(sys.exc_info()[1]))
